// Boost pinger (c) 2011 by Guilherme Maciel Ferreira / Intra2net AG
// Based upon work copyright (c) 2003-2010 Christopher M. Kohlhoff (ping.cpp)
// Modified 2014 by Christian Herdtweck, Intra2net AG
//
// Distributed under the Boost Software License, Version 1.0.
//    (See accompanying file LICENSE_1_0.txt or copy at
//          http://www.boost.org/LICENSE_1_0.txt)
#ifndef ICMP_PINGER_H
#define ICMP_PINGER_H

#include <stdint.h>

#include <string>

#include <boost/asio.hpp>
#include <boost/function.hpp>
#include <boost/shared_ptr.hpp>

#include "host/pinger.h"
#include "host/pingstatus.h"
#include "icmp/icmppacket.h"
#include "icmp/icmppacketdistributor.h"

class IcmpPinger;
typedef boost::shared_ptr<IcmpPinger> IcmpPingerItem;

//-----------------------------------------------------------------------------
// IcmpPinger
//-----------------------------------------------------------------------------

/**
 * @brief This class performs an ICMP ping to host using Boost Asio.
 * Scope: one object per host.
 */
class IcmpPinger : public Pinger
{
public:
    static PingerItem create(
            const IoServiceItem io_serv,
            const boost::asio::ip::icmp::socket::protocol_type &protocol,
            const std::string &source_network_interface,
            const int echo_reply_timeout_in_sec
    );

    virtual ~IcmpPinger();

    virtual void ping(
            const std::string &destination_ip,
            const uint16_t destination_port,
            boost::function<void(bool)> ping_done_callback
    );

    virtual void stop_pinging();

private:
    IcmpPinger(
            const IoServiceItem io_serv,
            const boost::asio::ip::icmp::socket::protocol_type &protocol,
            const int echo_reply_timeout_in_sec
    );

    void set_destination_endpoint( const std::string &destination_ip );

    bool start_send();
    bool send_echo_request( const IcmpPacketItem icmp_packet );
    void schedule_timeout_echo_reply();
    void handle_timeout(const boost::system::error_code &error);

    bool handle_receive_icmp_packet(const IcmpPacketItem icmp_packet,
                                    const size_t bytes_transferred
    );

    void set_ping_status( PingStatus ping_status );

private:
    IcmpPacketDistributorItem PacketDistributor;

    /// The destination host
    boost::asio::ip::icmp::endpoint DestinationEndpoint;
    /// Network layer protocol used to ping, IPv4 or IPv6
    boost::asio::ip::icmp::socket::protocol_type Protocol;
    /// The timer of ICMP packet receive, triggers the timeout to avoid infinite
    /// wait
    boost::asio::deadline_timer IcmpPacketReceiveTimer;
    /// ICMP packet identifier
    uint16_t Identifier;
    /// ICMP packet sequence_number
    uint16_t SequenceNumber;
    /// The time when the last ICMP packet was sent
    boost::posix_time::ptime TimeSent;
    /// Flag to indicate if we got a reply or not
    bool ReplyReceived;
    /// The amount of time to wait for the reply
    int EchoReplyTimeoutInSec;
    /// The status of the pinger
    PingStatus PingerStatus;
    /// Callback to notify when the ping is done (got reply/timeout)
    boost::function< void(bool) > PingDoneCallback;
};

#endif // ICMP_PINGER_H
